//================================================================================================
// $Id: parserext.cpp 7226 2024-01-11 19:24:54Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Parser extensions
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "parserext.h"
#include "version.h"
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <map>
#include <iomanip>

#ifdef _MSC_VER
#  pragma warning ( disable : 4996 )
#endif // _MSC_VER

namespace wtc
{

const std::string Options::BinaryExtension(".wtb");
const std::string Options::ImportExtension(".inc");

class File : public wtparser::IFile
{
	FILE* mFile;

public:
	File(FILE* file)
		: wtparser::IFile()
		, mFile(file)
	{
		// ok
	}

	~File()
	{
		fclose(mFile);
	}

	bool Eof() const
	{
		return !!feof(mFile);
	}

	unsigned int Read(void* buffer, unsigned int size)
	{
		return static_cast<unsigned int>(fread(buffer, 1, size, mFile));
	}
};

} // wtc

wtc::SourceManager::SourceManager() : wtparser::ISourceManager()
{
	// ok
}

wtparser::IFile* wtc::SourceManager::OpenFile(const std::string & filename)
{
	FILE* file = fopen(filename.c_str(), "rb");

	if(file)
	{
		wtc::File* ifile = new wtc::File(file);

		if(ifile)
		{
			return ifile;
		}
		else
		{
			fclose(file);
			return 0;
		}
	}
	else
	{
		return 0;
	}
}

wtc::ProblemHandler::ProblemHandler()
	: wtparser::IProblemHandler()
	, mNumNotes(0)
	, mNumWarnings(0)
	, mNumErrors(0)
{
	// ok
}

const std::string & wtc::ProblemHandler::Problem::KindToString(Kind kind)
{
	static const std::string k_unknown("<UNKNOWN>");
	static const std::string k_note("NOTE");
	static const std::string k_warning("WARNING");
	static const std::string k_error("ERROR");

	switch(kind)
	{
	case NOTE:    return k_note;
	case WARNING: return k_warning;
	case ERROR:   return k_error;
	default:      return k_unknown;
	}
}

std::string wtc::ProblemHandler::Problem::ToString() const
{
	std::ostringstream ss;
	ss << file << ":" << line << ": " << KindToString(kind) << ": " << msg;
	return ss.str();
}

void wtc::ProblemHandler::AddNote(const std::string & file, unsigned int line, const std::string & msg)
{
	++mNumNotes;
	mProblems.push_back(Problem(Problem::NOTE, file, line, msg));
}

void wtc::ProblemHandler::AddWarning(const std::string & file, unsigned int line, const std::string & msg)
{
	++mNumWarnings;
	mProblems.push_back(Problem(Problem::WARNING, file, line, msg));
}

void wtc::ProblemHandler::AddError(const std::string & file, unsigned int line, const std::string & msg)
{
	++mNumErrors;
	mProblems.push_back(Problem(Problem::ERROR, file, line, msg));
}

bool wtc::Options::IsHostBigEndian() const
{
	static union bint_t
	{ 
		unsigned int  i; 
		unsigned char b[4]; 
	} bint = { 0x00010203ul };
 
	return 0==bint.b[0];
}

void wtc::Options::AddCommand(CommandDescriptor desc)
{
	mCommands[desc.name] = desc;
}

void wtc::Options::AddCommands()
{
	std::string nul("");
	std::string def(" (default)");

	AddCommand(CommandDescriptor("--help", "Print help screen", &Options::CmdHelp));
	AddCommand(CommandDescriptor("--version", "Print version information", &Options::CmdVersion));
	AddCommand(CommandDescriptor("-dump-ast", "Dump generated AST", &Options::CmdDumpAst));
	AddCommand(CommandDescriptor("-syntax-only", "Check syntax only, do not generate output", &Options::CmdSyntaxOnly));
	AddCommand(CommandDescriptor("-debug-lexer", "Enable lexer debugging", &Options::CmdDebugLexer));
	AddCommand(CommandDescriptor("-debug-rules", "Enable rule debugging", &Options::CmdDebugRules));
	AddCommand(CommandDescriptor("-o", "Set output file name", &Options::CmdOutput));
	AddCommand(CommandDescriptor("-labels-none", "Do not add any labels into binary", &Options::CmdLabelsNone));
	AddCommand(CommandDescriptor("-labels-top", "Add top-level labels into binary", &Options::CmdLabelsTop));
	AddCommand(CommandDescriptor("-labels-all", "Add all labels into binary", &Options::CmdLabelsAll));
	AddCommand(CommandDescriptor("-big-endian", std::string("Write binary in big endian") + (IsHostBigEndian()?def:nul), &Options::CmdBigEndian));
	AddCommand(CommandDescriptor("-little-endian", std::string("Write binary in little endian") + (IsHostBigEndian()?nul:def), &Options::CmdLittleEndian));
	AddCommand(CommandDescriptor("-inline-32", "Inline all 32 bit constants", &Options::CmdInline32));
	AddCommand(CommandDescriptor("-inline-64", "Inline all 64 bit constants", &Options::CmdInline64));
	AddCommand(CommandDescriptor("-no-inline-32", "Disable inlining of 32 bit constants", &Options::CmdNoInline32));
	AddCommand(CommandDescriptor("-no-inline-64", "Disable inlining of 64 bit constants", &Options::CmdNoInline64));
	AddCommand(CommandDescriptor("-expect-error", "Expect error during compilation (swap exit code)", &Options::CmdExpectError));
	AddCommand(CommandDescriptor("-disable-asserts", "Disable code generation of assertions", &Options::CmdDisableAsserts));
	AddCommand(CommandDescriptor("-disable-precond", "Disable code generation of precondition", &Options::CmdDisablePrecond));
	AddCommand(CommandDescriptor("-disable-postcond", "Disable code generation of postcondition", &Options::CmdDisablePostcond));
	AddCommand(CommandDescriptor("-gen-import", "Generate import include from exports", &Options::CmdGenerateImportInclude));
	AddCommand(CommandDescriptor("-impinc-name", "Set file name of generated import include", &Options::CmdImportIncludeName));
	AddCommand(CommandDescriptor("-add-hash", "Add an integrity section using the desired hash algorithm", &Options::CmdAddHash));
	AddCommand(CommandDescriptor("-signed-index", "Enable the usage of signed integers as indices", &Options::CmdSignedIndex));
	AddCommand(CommandDescriptor("-no-signed-index", "Disable the usage of signed integers as indices", &Options::CmdNoSignedIndex));
	AddCommand(CommandDescriptor("-large-index", "Enable the usage of 64-bit integers as indices", &Options::CmdLargeIndex));
	AddCommand(CommandDescriptor("-no-large-index", "Disable the usage of 64-bit integers as indices", &Options::CmdNoLargeIndex));
}

wtc::Options::Options(int argc, char** argv)
	: wtparser::IOptions()
	, mProgramName(argv[0])
	, mDebugLexer(false)
	, mDebugRules(false)
	, mAllowFloatingBlocks(true)
	, mDumpAst(false)
	, mSyntaxOnly(false)
	, mLabelsTop(false)
	, mLabelsAll(false)
	, mInline32(true)
	, mInline64(true)
	, mBigEndian(IsHostBigEndian())
	, mExpectError(false)
	, mDisableAsserts(false)
	, mDisablePrecond(false)
	, mDisablePostcond(false)
	, mGenerateImportInclude(false)
	, mSignedIndex(false)
	, mLargeIndex(false)
	, mSourceFilename()
	, mOutputFilename()
	, mImportFilename()
{
	AddCommands();

	static const int arg_offset = 1;
	argv += arg_offset;
	argc -= arg_offset;

	while(argc>0)
	{
		if('-'==(*argv)[0])
		{
			commands_t::iterator it = mCommands.find(*argv);

			if(mCommands.end()!=it)
			{
				++argv;
				--argc;
				(this->*it->second.cmd)(argc, argv);

				if(!mProblemString.empty())
				{
					return;
				}
			}
			else
			{
				std::ostringstream ss;
				ss << "Unknown operand: \"" << (*argv) << "\".";
				mProblemString = ss.str();
				return;
			}
		}
		else
		if(mSourceFilename.empty())
		{
			mSourceFilename = *argv;
			++argv;
			--argc;
		}
		else
		{
			mProblemString = "Multiple file name operand. Only one is allowed.";
			return;
		}
	}

	if(mSourceFilename.empty())
	{
		mProblemString = "Missing file name operand.";
		return;
	}

	if(mOutputFilename.empty())
	{
		mOutputFilename = ChangeWTExtensionTo(mSourceFilename, BinaryExtension);
	}

	if(mImportFilename.empty())
	{
		mImportFilename = ChangeWTExtensionTo(mSourceFilename, ImportExtension);
	}
}

std::string wtc::Options::ChangeWTExtensionTo(const std::string & str, const std::string & ext)
{
	std::size_t len = str.length();
	std::size_t pos = len - 3;

	if(len>=4 && '.'==str[pos+0] && 'w'==str[pos+1] && 't'==str[pos+2])
	{
		return str.substr(0, pos) + ext;
	}
	else
	{
		return str + ext;
	}
}

void wtc::Options::CmdHelp(int & argc, char** & argv)
{
	std::cout << "Usage: " << mProgramName << " [OPTION]... FILE" << std::endl;

	std::cout << "Welltype compiler v"
		<< wtc::version::Major << '.'
		<< wtc::version::Minor << '.'
		<< wtc::version::Patch
		<< std::endl;

	std::cout << std::endl;

	std::size_t width = 0;

	for(commands_t::iterator it=mCommands.begin(),end=mCommands.end();it!=end;++it)
	{
		if(width<it->second.name.length())
		{
			width = it->second.name.length();
		}
	}

	width += 2;

	for(commands_t::iterator it=mCommands.begin(),end=mCommands.end();it!=end;++it)
	{
		const CommandDescriptor & desc = it->second;
		std::cout << "  " << std::setw(width) << std::left << desc.name << std::resetiosflags(std::ios::showbase) << desc.help << std::endl;
	}

	exit(0);
}

void wtc::Options::CmdVersion(int & argc, char** & argv)
{
	std::cout << "Welltype compiler v"
		<< wtc::version::Major << '.'
		<< wtc::version::Minor << '.'
		<< wtc::version::Patch
		<< std::endl;

	exit(0);
}

void wtc::Options::CmdDumpAst(int & argc, char** & argv)
{
	mDumpAst = true;
}

void wtc::Options::CmdSyntaxOnly(int & argc, char** & argv)
{
	mSyntaxOnly = true;
}

void wtc::Options::CmdDebugLexer(int & argc, char** & argv)
{
	mDebugLexer = true;
}

void wtc::Options::CmdDebugRules(int & argc, char** & argv)
{
	mDebugRules = true;
}

void wtc::Options::CmdOutput(int & argc, char** & argv)
{
	if(argc<2)
	{
		mProblemString = "Not enough parameters.";
		return;
	}

	mOutputFilename = *argv;
	++argv;
	--argc;
}

void wtc::Options::CmdLabelsNone(int & argc, char** & argv)
{
	mLabelsTop = false;
	mLabelsAll = false;
}

void wtc::Options::CmdLabelsTop(int & argc, char** & argv)
{
	mLabelsTop = true;
}

void wtc::Options::CmdLabelsAll(int & argc, char** & argv)
{
	mLabelsTop = true;
	mLabelsAll = true;
}

void wtc::Options::CmdBigEndian(int & argc, char** & argv)
{
	mBigEndian = true;
}

void wtc::Options::CmdLittleEndian(int & argc, char** & argv)
{
	mBigEndian = false;
}

void wtc::Options::CmdInline32(int & argc, char** & argv)
{
	mInline32 = true;
}

void wtc::Options::CmdInline64(int & argc, char** & argv)
{
	mInline64 = true;
}

void wtc::Options::CmdNoInline32(int & argc, char** & argv)
{
	mInline32 = false;
}

void wtc::Options::CmdNoInline64(int & argc, char** & argv)
{
	mInline64 = false;
}

void wtc::Options::CmdExpectError(int & argc, char** & argv)
{
	mExpectError = true;
}

void wtc::Options::CmdDisableAsserts(int & argc, char** & argv)
{
	mDisableAsserts = true;
}

void wtc::Options::CmdDisablePrecond(int & argc, char** & argv)
{
	mDisablePrecond = true;
}

void wtc::Options::CmdDisablePostcond(int & argc, char** & argv)
{
	mDisablePostcond = true;
}

void wtc::Options::CmdGenerateImportInclude(int & argc, char** & argv)
{
	mGenerateImportInclude = true;
}

void wtc::Options::CmdImportIncludeName(int & argc, char** & argv)
{
	if(argc<2)
	{
		mProblemString = "Not enough parameters.";
		return;
	}

	mImportFilename = *argv;
	++argv;
	--argc;
}

void wtc::Options::CmdAddHash(int & argc, char** & argv)
{
	if(argc<2)
	{
		mProblemString = "Not enough parameters.";
		return;
	}

	mHashAlgs.insert(*argv);
	++argv;
	--argc;
}

void wtc::Options::CmdSignedIndex(int & argc, char** & argv)
{
	mSignedIndex = true;
}

void wtc::Options::CmdNoSignedIndex(int & argc, char** & argv)
{
	mSignedIndex = false;
}

void wtc::Options::CmdLargeIndex(int & argc, char** & argv)
{
	mLargeIndex = true;
}

void wtc::Options::CmdNoLargeIndex(int & argc, char** & argv)
{
	mLargeIndex = false;
}
