#include "config_private.h"
#include "dynstack.h"
#include "vm.h"
#include <string.h>

wt_bool wtr_dynstack_init(struct wtr_dynstack* const stk,
	struct wtr_vm* const vm, wt_u32 minsize, wt_u32 maxsize, wt_u32 limit)
{
	stk->sp       = 0;
	stk->size     = minsize;
	stk->min_size = minsize;
	stk->max_size = maxsize;
	stk->limit    = limit;
	stk->vm       = vm;

	if(NULL==(stk->data=(struct wtr_objentry*)wtr_vm_sys_malloc(vm,
		sizeof(struct wtr_objentry)*minsize)))
	{
		return WT_FALSE;
	}

	return WT_TRUE;
}

void wtr_dynstack_free(struct wtr_dynstack* const stk)
{
	if(NULL!=stk->data)
	{
		wtr_vm_sys_free(stk->vm, stk->data);
		stk->data = NULL;
	}
}

static wt_bool enlarge_on_demand(struct wtr_dynstack* const stk)
{
	if(stk->sp<stk->size)
	{
		/* branch prediction optimization */
	}
	else
	{
		const size_t raw_size = sizeof(struct wtr_objentry)*stk->size;

		if(stk->size==stk->max_size)
		{
			return 1;
		}

		struct wtr_objentry* const new_arr =
			(struct wtr_objentry*)wtr_vm_sys_realloc(stk->vm,
				stk->data, raw_size, raw_size<<1);

		if(NULL==new_arr)
		{
			return 2;
		}

		stk->data    = new_arr;
		stk->size  <<= 1;
		stk->limit <<= 1;
	}

	return 0;
}

struct wtr_objentry* wtr_dynstack_push_p(struct wtr_dynstack* const stk,
	const struct wtr_objentry* const ent)
{
	struct wtr_objentry* top;

	if(0!=enlarge_on_demand(stk))
	{
		return NULL;
	}

	top = stk->data + stk->sp;
	memcpy(top, ent, sizeof(struct wtr_objentry));
	++stk->sp;

	return top;
}

struct wtr_objentry* wtr_dynstack_push_v(struct wtr_dynstack* const stk,
	wt_vmptr entry_vma)
{
	struct wtr_objentry* top;

	if(0!=enlarge_on_demand(stk))
	{
		return NULL;
	}

	top = stk->data + stk->sp;

	if(0!=wtr_vma_read(stk->vm, entry_vma, top, sizeof(struct wtr_objentry)))
	{
		return NULL;
	}

	++stk->sp;

	return top;
}

struct wtr_objentry* wtr_dynstack_top(struct wtr_dynstack* const stk)
{
	if(0<stk->sp)
	{
		return stk->data + stk->sp - 1;
	}

	return NULL;
}

wt_bool wtr_dynstack_pop(struct wtr_dynstack* const stk)
{
	if(stk->sp)
	{
		--stk->sp;

		if(stk->sp>=stk->limit || stk->size<=stk->min_size)
		{
			return WT_TRUE;
		}
		else
		{
			const size_t raw_size = sizeof(struct wtr_objentry)*stk->size;

			struct wtr_objentry* const new_arr =
				(struct wtr_objentry*)wtr_vm_sys_realloc(stk->vm,
					stk->data, raw_size, raw_size/2);

			if(NULL!=new_arr)
			{
				stk->data    = new_arr;
				stk->size  >>= 1;
				stk->limit >>= 1;
			}

			return WT_TRUE;
		}
	}

	return WT_FALSE;
}
