#include "config_private.h"
#include "gidtree.h"
#include "vm.h"
#include <string.h>
#if WTRUNTIME_DEBUG
#  include <assert.h>
#endif /* WTRUNTIME_DEBUG */

#if 0<WTRUNTIME_GIDTREE_NODEBITS
#  define NODEBITS WTRUNTIME_GIDTREE_NODEBITS
#else /* 0<WTRUNTIME_GIDTREE_NODEBITS */
/* 64 entries per node (+ size of a pointer) */
#  define NODEBITS (6+WTRUNTIME_PTR_BITS)
#endif /* 0<WTRUNTIME_GIDTREE_NODEBITS */

#if NODEBITS<=WTRUNTIME_PTR_BITS
#  error "WTRUNTIME_GIDTREE_NODEBITS is too small"
#endif /* NODEBITS<=WTRUNTIME_PTR_BITS */

#define NODESIZE    (1<<(NODEBITS))
#define ENTS_BITS   (NODEBITS-WTRUNTIME_PTR_BITS)
#define ENTS_P_N    (1<<ENTS_BITS) /* entries per node */
#define FULLFLAG    1

/* maximal number of GIDs are limited to what a typeinfo can hold */
#define GID_BITS (32-WTRT_RES_TYPEBITS)

#define MAX_LEVELS \
	((GID_BITS / ENTS_BITS) + ((0!=(GID_BITS % ENTS_BITS)) ? 1 : 0))

#define NODEPTR(tab) \
	((void**)(((wt_uintptr_t)(tab)) & (~(wt_uintptr_t)FULLFLAG)))

#define GID_COUNT(levels) \
	(1<<(((levels)+1)*ENTS_BITS))

int wtr_gidtree_create(struct wtr_gidtree* const gidtree,
	struct wtr_vm* const vm)
{
	gidtree->levels = 0;

	if(NULL==(gidtree->root=(void**)wtr_vm_sys_malloc(vm, NODESIZE)))
	{
		return 1;
	}

	memset(gidtree->root, 0, NODESIZE);

	return 0;
}

static void release_node_rec(struct wtr_vm* const vm, void** const node,
	size_t levels)
{
	if(NULL==node)
	{
		return;
	}

	if(0<(levels--))
	{
		size_t i;

		for(i=0;i<ENTS_P_N;++i)
		{
			release_node_rec(vm, NODEPTR(node[i]), levels);
		}
	}

	wtr_vm_sys_free(vm, node);
}

void wtr_gidtree_release(struct wtr_gidtree* const gidtree,
	struct wtr_vm* const vm)
{
	if(NULL!=gidtree->root)
	{
		release_node_rec(vm, NODEPTR(gidtree->root), gidtree->levels);
		memset(gidtree, 0, sizeof(*gidtree));
	}
}

/*
	returns:
	   0  on success
	  +1 on full
	  -1 on allocation error
*/
static int add_sym(struct wtr_vm* const vm, void** const node,
	const size_t level, wt_globalid* const out_gid, struct wtr_sym* const sym)
{
	size_t i;

	if(0<level)
	{
		const size_t next_level = level - 1;
		long empty = -1;

		for(i=0;i<ENTS_P_N;++i)
		{
			wt_uintptr_t sub = (wt_uintptr_t)node[i];

			if(0==sub)
			{
				if(0>empty)
				{
					empty = i;
				}
			}
			else
			if(!(FULLFLAG & sub))
			{
				const int res =
					add_sym(vm, (void**)sub, next_level, out_gid, sym);

				if(0==res)
				{
					*out_gid |= i << (level * ENTS_BITS);
					return 0;
				}
				else
				if(0>res)
				{
					return res;
				}
				else
				{
					node[i] = (void*)(FULLFLAG | sub);
				}
			}
		}

		if(0<=empty)
		{
			if(NULL==(node[empty]=(void**)wtr_vm_sys_malloc(vm, NODESIZE)))
			{
				return -1;
			}

			memset(node[empty], 0, NODESIZE);

			{
				const int res =
					add_sym(vm, (void**)node[empty], next_level, out_gid, sym);

				if(0==res)
				{
					*out_gid |= empty << (level * ENTS_BITS);
					return 0;
				}
				else
				{
#if WTRUNTIME_DEBUG
					assert(0>res);
#endif /* WTRUNTIME_DEBUG */
					return res;
				}
			}
		}
	}
	else
	{
		for(i=0;i<ENTS_P_N;++i)
		{
			if(NULL==node[i])
			{
				node[i] = sym;
				*out_gid = i;
				return 0;
			}
		}
	}

	return 1;
}

static void** get_gid_slot(struct wtr_gidtree* const gidtree, wt_globalid gid,
	const int clear)
{
	size_t levels = gidtree->levels;

	if(GID_COUNT(levels)>gid)
	{
		void** node = gidtree->root;

		while(0<levels)
		{
			size_t idx = (gid >> (ENTS_BITS*levels)) % ENTS_P_N;

			--levels;

			if(clear)
			{
				node[idx] = (void*)(
					((wt_uintptr_t)node[idx]) & (~(wt_uintptr_t)FULLFLAG)
					);
			}

			node = NODEPTR(node[idx]);

			if(NULL==node)
			{
				return NULL;
			}
		}

		return &node[gid % ENTS_P_N];
	}

	return NULL;
}

wt_error wtr_gidtree_add(struct wtr_gidtree* const gidtree,
	struct wtr_vm* const vm, wt_globalid* const out_gid,
	struct wtr_sym* const sym)
{
	int res;
	int retry = 1;

again:
	res = add_sym(vm, gidtree->root, gidtree->levels, out_gid, sym);

	if(0==res)
	{
		wt_globalid gid = *out_gid;

		if((1<<GID_BITS)<=gid)
		{
			*get_gid_slot(gidtree, gid, 1) = NULL;
			return WTE_GIDFULL;
		}

		return WTE_OK;
	}
	else
	if(0>res)
	{
		return WTE_NOMEM;
	}

	if(retry && MAX_LEVELS>gidtree->levels)
	{
		/* add a new root, to extend the tree by 1 level */

		void** const new_root = (void**)wtr_vm_sys_malloc(vm, NODESIZE);

		if(NULL==new_root)
		{
			return WTE_NOMEM;
		}

		memset(new_root, 0, NODESIZE);

		new_root[0] = gidtree->root;
		gidtree->root = new_root;
		++gidtree->levels;

		retry = 0;
		goto again;
	}

	return WTE_GIDFULL;
}

void wtr_gidtree_clear(struct wtr_gidtree* const gidtree, wt_globalid gid)
{
	void** const slot = get_gid_slot(gidtree, gid, 1);

	if(NULL!=slot)
	{
		*slot = NULL;
	}
}

struct wtr_sym* wtr_gidtree_get(struct wtr_gidtree* const gidtree,
	wt_globalid gid)
{
	void** const slot = get_gid_slot(gidtree, gid, 0);

	if(NULL!=slot)
	{
		return (struct wtr_sym*)(*slot);
	}

	return NULL;
}

static int find_gid_rec(void** const node, const size_t level,
	wt_globalid* const out_gid, void* const ctx, int (*test)(void* const ctx,
	struct wtr_sym* const sym))
{
	size_t i;

	if(0<level)
	{
		const size_t next_level = level - 1;

		for(i=0;i<ENTS_P_N;++i)
		{
			wt_uintptr_t sub = (wt_uintptr_t)node[i];

			if(0!=sub)
			{
				const int res = find_gid_rec(NODEPTR(sub), next_level,
					out_gid, ctx, test);

				if(0==res)
				{
					*out_gid |= i << (level * ENTS_BITS);
					return 0;
				}
			}
		}
	}
	else
	{
		for(i=0;i<ENTS_P_N;++i)
		{
			if(NULL!=node[i] && 0==test(ctx, (struct wtr_sym*)node[i]))
			{
				*out_gid = i;
				return 0;
			}
		}
	}

	return 1;
}

int wtr_gidtree_find(struct wtr_gidtree* const gidtree,
	wt_globalid* const out_gid, void* const ctx, int (*test)(void* const ctx,
	struct wtr_sym* const sym))
{
	if(NULL!=gidtree->root)
	{
		return find_gid_rec(NODEPTR(gidtree->root), gidtree->levels,
			out_gid, ctx, test);
	}

	return 1;
}
