#include "config_private.h"
#include "hash.h"
#include "byteorder.h"
#include "wtruntime.h"
#include "lreader.h"
#include "../welltype/wtb3.h"
#include <string.h>

const struct wtr_hashalg* wtr_hash_find_by_name(
	const struct wtr_hashalg* algs, const char* const name)
{
	if(NULL==algs)
	{
		return NULL;
	}

	while(NULL!=algs->name)
	{
		if(0==strcmp(name, algs->name))
		{
			return algs;
		}

		++algs;
	}

	return NULL;
}

const struct wtr_hashalg* wtr_hash_find_by_id(
	const struct wtr_hashalg* algs, const wt_u16 id)
{
	if(NULL==algs)
	{
		return NULL;
	}

	while(NULL!=algs->name)
	{
		if(id==algs->hashalg_id)
		{
			return algs;
		}

		++algs;
	}

	return NULL;
}

static wt_error wtr_hash_update__(const struct wtr_hashalg* const alg,
	void* const ctx, struct wtr_lreader* const lr, wt_uint32_t offs,
	wt_uint32_t size)
{
	wt_uint8_t buf[32];

	while(sizeof(buf)<size)
	{
		if(sizeof(buf)!=lr->read_sys(lr, buf, offs, sizeof(buf)))
		{
			return WTE_FAULT;
		}

		alg->update(alg, ctx, buf, sizeof(buf));
	}

	if(size!=lr->read_sys(lr, buf, offs, size))
	{
		return WTE_FAULT;
	}

	alg->update(alg, ctx, buf, size);

	return WTE_OK;
}

#define ALG_UPDATE(_offs_, _size_) \
	do \
	{ \
		const wt_error err = wtr_hash_update__(alg, ctx, lr, (_offs_), \
			(_size_)); \
		\
		if(WTE_OK!=err) \
		{ \
			return err; \
		} \
	} \
	while(0)

static wt_error wtr_hash_visitor__(const struct wtr_hashalg* const alg,
	void* const ctx, struct wtr_lreader* const lr)
{
	if(sizeof(struct wtb3_hdr)>lr->size)
	{
		return WTE_BINERR;
	}

	if(WTB3MAG0!=lr->read8(lr,
		offsetof(struct wtb3_hdr, b_ident)+WTB3I_MAGIC0))
	{
		return WTE_BINERR;
	}

	if(WTB3MAG1!=lr->read8(lr,
		offsetof(struct wtb3_hdr, b_ident)+WTB3I_MAGIC1))
	{
		return WTE_BINERR;
	}

	if(WTB3MAG2!=lr->read8(lr,
		offsetof(struct wtb3_hdr, b_ident)+WTB3I_MAGIC2))
	{
		return WTE_BINERR;
	}

	if(WTB3MAG3!=lr->read8(lr,
		offsetof(struct wtb3_hdr, b_ident)+WTB3I_MAGIC3))
	{
		return WTE_BINERR;
	}

	if(WTB3V_CURRENT!=lr->read8(lr,
		offsetof(struct wtb3_hdr, b_ident)+WTB3I_VERSION))
	{
		return WTE_BINERR;
	}

	/* header */
	{
		ALG_UPDATE(0, sizeof(struct wtb3_hdr));
	}

	/* section headers */
	{
		const wt_uint16_t b_shnum =
			lr->read16(lr, offsetof(struct wtb3_hdr, b_shnum));

		const wt_uint32_t b_shoff =
			lr->read32(lr, offsetof(struct wtb3_hdr, b_shoff));

		ALG_UPDATE(b_shoff, sizeof(struct wtb3_shdr)*b_shnum);
	}

	/* thread headers */
	{
		const wt_uint16_t b_thnum =
			lr->read16(lr, offsetof(struct wtb3_hdr, b_thnum));

		const wt_uint32_t b_thoff =
			lr->read32(lr, offsetof(struct wtb3_hdr, b_thoff));

		ALG_UPDATE(b_thoff, sizeof(struct wtb3_thdr)*b_thnum);
	}

	/* section data */
	{
		const wt_uint16_t b_shnum =
			lr->read16(lr, offsetof(struct wtb3_hdr, b_shnum));

		wt_uint32_t b_shoff =
			lr->read32(lr, offsetof(struct wtb3_hdr, b_shoff));

		wt_uint16_t i;

		for(i=1;i<b_shnum;b_shoff+=sizeof(struct wtb3_shdr),++i)
		{
			const wt_uint8_t sh_type = lr->read8(lr,
				b_shoff+offsetof(struct wtb3_shdr, sh_type));
			const wt_uint8_t sh_flag = lr->read8(lr,
				b_shoff+offsetof(struct wtb3_shdr, sh_flag));
			const wt_uint32_t sh_size = lr->read32(lr,
				b_shoff+offsetof(struct wtb3_shdr, sh_size));
			const wt_uint32_t sh_offs = lr->read32(lr,
				b_shoff+offsetof(struct wtb3_shdr, sh_offs));

			/* do not hash 0 length data, because it does not any sense */
			if(0==sh_size)
			{
				continue;
			}

			/* CLEAR and LOAD flags excludes themselves */
			if((WTB3SF_CLEAR & sh_flag) && (WTB3SF_LOAD & sh_flag))
			{
				return WTE_BINERR;
			}

			/* if the CLEAR flag is set, it means no associated data */
			if(WTB3SF_CLEAR & sh_flag)
			{
				continue;
			}

			/* INTEGRITY sections: we can hash only the headers, and we should do it */
			/* -- otherwise problems will raise to find the fix-point of the hash value */
			if(WTB3ST_INTEGRITY==sh_type)
			{
				ALG_UPDATE(sh_offs, sizeof(struct wtb3_integhdr));
				continue;
			}

			/* section data is ready to be hashed */
			ALG_UPDATE(sh_offs, sh_size);
		}
	}

	return WTE_OK;
}

wt_error wtr_hash_calc(void* out_digest, const struct wtr_hashalg* const alg,
	const void* data, size_t size, wt_bo bo)
{
	struct wtr_lreader lr;

	/* pass NULL as vm, but this is not a problem, this code will not use */
	/* anything that really requires it */
	wtr_lreader_init_sys(&lr, NULL, data, size, bo);

	return wtr_hash_calc_lr(out_digest, &lr, alg);
}

wt_error wtr_hash_calc_lr(void* out_digest, struct wtr_lreader* const lr,
	const struct wtr_hashalg* const alg)
{
	wt_error err = WTE_OK;

	void* ctx;

	if(0!=alg->init(alg, &ctx))
	{
		return WTE_HAINIT;
	}

	err = wtr_hash_visitor__(alg, ctx, lr);

	alg->final(alg, ctx, out_digest);

	return err;
}

wt_error wtr_hash_check(const void* digest,
	const struct wtr_hashalg* const alg, const void* data, size_t size,
	wt_bo bo)
{
	struct wtr_lreader lr;

	/* pass NULL as vm, but this is not a problem, this code will not use */
	/* anything that really requires it */
	wtr_lreader_init_sys(&lr, NULL, data, size, bo);

	return wtr_hash_check_lr(&lr, alg,
		((const wt_uint8_t*)digest)-((const wt_uint8_t*)data));
}

wt_error wtr_hash_check_lr(struct wtr_lreader* const lr,
	const struct wtr_hashalg* const alg, const wt_uint32_t digest_offs)
{
	wt_u8    calc_digest[WTR_MAX_HASH_DIGEST_LEN];
	wt_u8    ref_digest[WTR_MAX_HASH_DIGEST_LEN];
	wt_error err;

	if(0==alg->digest_len || WTR_MAX_HASH_DIGEST_LEN<alg->digest_len)
	{
		return WTE_HADIGL;
	}

	if(alg->digest_len!=lr->read_sys(lr, ref_digest, digest_offs,
		alg->digest_len))
	{
		return WTE_FAULT;
	}

	if(WTE_OK!=(err=wtr_hash_calc_lr(calc_digest, lr, alg)))
	{
		return err;
	}

	if(0!=memcmp(ref_digest, calc_digest, alg->digest_len))
	{
		return WTE_INTEG;
	}

	return WTE_OK;
}
